/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * include/SIL/time.h: Header for timekeeping routines.
 */

/*
 * In addition to basic timekeeping functions (obtain the current time,
 * wait for a period of time), this header also provides a simple "marker"
 * interface that can be used for low-granularity profiling, by recording
 * the times at which particular events occurred and later reading out
 * those times as offsets from a fixed initial time.  For example, the
 * caller might call time_mark_reset() at the beginning of a frame, then
 * call time_mark() at certain points during the frame -- after processing
 * events, when rendering completes, and so on.  The caller could then use
 * the values returned by time_get_mark() in drawing a profiling display.
 *
 * Marker handling is automatically disabled if DEBUG is not defined;
 * time_mark_reset() and time_mark() become no-ops, and time_get_mark()
 * always returns a negative value (indicating a nonexistent mark).
 */

#ifndef SIL_TIME_H
#define SIL_TIME_H

EXTERN_C_BEGIN

/*************************************************************************/
/************************* Configuration options *************************/
/*************************************************************************/

/**
 * TIME_MAX_MARKS:  Maximum number of marks that can be registered with
 * time_mark() for a single call to time_mark_reset().
 */
#define TIME_MAX_MARKS  64

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

/*---------------------------- Core routines ----------------------------*/

/**
 * time_now:  Return the current time as a scalar value.  The return value
 * is in units of seconds and increases monotonically over the course of
 * the program's execution.  The epoch is no later than the starting time
 * of the program, so timestamp values returned by this function are always
 * nonnegative; however, the epoch is otherwise undefined, so the actual
 * value of a timestamp is meaningless except in comparison to other
 * timestamps.
 *
 * If the system time service provides subsecond resolution, this function
 * will also return timestamps with subsecond resolution.
 *
 * [Return value]
 *     Current timestamp, in seconds.
 */
extern double time_now(void);

/**
 * time_delay:  Wait for the specified amount of time.  (Fractional values
 * are allowed.)  Returns immediately if the specified time is zero or
 * negative.
 *
 * [Parameters]
 *     time: Amount of time to delay, in seconds.
 */
extern void time_delay(double time);

/*-------------------- Time markers (for debugging) ---------------------*/

/**
 * time_mark_reset:  Set the current time as the reference time for calls
 * to time_get_mark(), and clear all previously set marks.
 *
 * This function is a no-op if DEBUG is not defined.
 */
#ifdef DEBUG
extern void time_mark_reset(void);
#else
#define time_mark_reset()  /*nothing*/
#endif

/**
 * time_mark:  Associate the current time with the given mark value, for
 * later retrieval with time_get_mark().  Behavior is undefined if the
 * same mark value is used multiple times with no intervening call to
 * time_mark_reset().
 *
 * The number of marks that can be recorded is limited by the
 * TIME_MAX_MARKS constant defined in src/time.c.
 *
 * This function is a no-op if DEBUG is not defined.
 *
 * [Parameters]
 *     mark: Mark value to associate current time with.
 */
#ifdef DEBUG
extern void time_mark(int mark);
#else
#define time_mark(mark)  /*nothing*/
#endif

/**
 * time_get_mark:  Retrieve the time associated with the given mark value
 * as the number of seconds elapsed between the time_mark_reset() call and
 * the time_mark() call for that mark value.  If the given mark value has
 * not been registered, an unspecified negative value is returned.
 *
 * This function always returns a negative value if DEBUG is not defined.
 *
 * [Parameters]
 *     mark: Mark value for which to retrieve elapsed time.
 * [Return value]
 *     Elapsed time in seconds, or a negative value if the mark does not exist.
 */
#ifdef DEBUG
extern double time_get_mark(int mark);
#else
#define time_get_mark(mark)  (-1)
#endif

/*************************************************************************/
/*************************************************************************/

EXTERN_C_END

#endif  // SIL_TIME_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
