/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/graphics/misc.c: Miscellaneous tests for graphics functions.
 */

#include "src/base.h"
#include "src/graphics.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"
#include "src/texture.h"

#if defined(SIL_PLATFORM_LINUX)
# include "src/sysdep/linux/internal.h"
# include <dlfcn.h>
# define RTLD_DEFAULT  NULL  // See note in src/sysdep/linux/graphics.c
# include <X11/Xlib.h>
# include <X11/extensions/Xrandr.h>
# include <X11/extensions/xf86vmode.h>
#elif defined(SIL_PLATFORM_MACOSX)
# include "src/sysdep/macosx/graphics.h"
# include <CoreGraphics/CoreGraphics.h>
#elif defined(SIL_PLATFORM_WINDOWS)
# include "src/sysdep/windows/internal.h"
#endif

/*************************************************************************/
/**************************** Helper routines ****************************/
/*************************************************************************/

/**
 * get_display_resolution:  Get the current display resolution from the
 * operating system.
 *
 * [Parameters]
 *     width_ret, height_ret: Pointers to variables to receive the current
 *         display resolution.
 * [Return value]
 *     True on success, false if the function is not implemented on the
 *     current platform.
 */
static int get_display_resolution(int *width_ret, int *height_ret)
{
#if defined(SIL_PLATFORM_LINUX)

    *width_ret = *height_ret = -1;

    #define LOOKUP(sym) \
        __typeof__(sym) *p_##sym; \
        p_##sym = dlsym(RTLD_DEFAULT, #sym)
    LOOKUP(XRRFreeCrtcInfo);
    LOOKUP(XRRFreeOutputInfo);
    LOOKUP(XRRFreeScreenResources);
    LOOKUP(XRRGetCrtcInfo);
    LOOKUP(XRRGetOutputInfo);
    LOOKUP(XRRGetScreenResources);
    LOOKUP(XRRQueryExtension);
    LOOKUP(XRRQueryVersion);
    LOOKUP(XF86VidModeGetModeLine);
    LOOKUP(XF86VidModeQueryExtension);
    #undef LOOKUP

    Display *display = linux_x11_display();
    int major, minor;
    if (p_XRRQueryExtension
     && (*p_XRRQueryExtension)(display, (int[1]){}, (int[1]){})
     && (*p_XRRQueryVersion)(display, &major, &minor)
     && (major >= 2 || (major == 1 && minor >= 2))) {
        XRRScreenResources *sr =
            (*p_XRRGetScreenResources)(display, RootWindow(display, 0));
        ASSERT(sr);
        for (int i = 0; i < sr->noutput; i++) {
            XRROutputInfo *oi =
                (*p_XRRGetOutputInfo)(display, sr, sr->outputs[i]);
            ASSERT(oi);
            if (!oi) {
                continue;
            }
            if (!oi->crtc || oi->connection == RR_Disconnected) {
                (*p_XRRFreeOutputInfo)(oi);
                continue;
            }
            XRRCrtcInfo *crtc = (*p_XRRGetCrtcInfo)(display, sr, oi->crtc);
            ASSERT(crtc);
            const int rotated =
                (crtc->rotation & (RR_Rotate_90 | RR_Rotate_270)) != 0;
            for (int j = 0; j < oi->nmode; j++) {
                const XRRModeInfo *mode = NULL;
                for (int k = 0; k < sr->nmode; k++) {
                    if (sr->modes[k].id == oi->modes[j]) {
                        mode = &sr->modes[k];
                        break;
                    }
                }
                ASSERT(mode);
                if (mode->id == crtc->mode) {
                    if (rotated) {
                        *width_ret = mode->height;
                        *height_ret = mode->width;
                    } else {
                        *width_ret = mode->width;
                        *height_ret = mode->height;
                    }
                    break;
                }
            }
            (*p_XRRFreeCrtcInfo)(crtc);
            (*p_XRRFreeOutputInfo)(oi);
            break;
        }
        (*p_XRRFreeScreenResources)(sr);
    } else if (p_XF86VidModeQueryExtension
            && (*p_XF86VidModeQueryExtension)(display,
                                              (int[1]){}, (int[1]){})) {
        int dotclock;
        XF86VidModeModeLine modeline;
        ASSERT((*p_XF86VidModeGetModeLine)(display, DefaultScreen(display),
                                           &dotclock, &modeline));
        *width_ret = modeline.hdisplay;
        *height_ret = modeline.vdisplay;
    } else {
        FAIL("No video mode interface found!");
    }

    ASSERT(*width_ret != -1);

#elif defined(SIL_PLATFORM_MACOSX)

    CGDirectDisplayID display_id = macosx_display_id(0);
    CGDisplayModeRef current_mode = CGDisplayCopyDisplayMode(display_id);
    *width_ret = CGDisplayModeGetWidth(current_mode);
    *height_ret = CGDisplayModeGetHeight(current_mode);

#elif defined(SIL_PLATFORM_WINDOWS)

    *width_ret = *height_ret = -1;

    for (int device = 0; ; device++) {
        DISPLAY_DEVICE device_info = {.cb = sizeof(device_info)};
        ASSERT(EnumDisplayDevices(NULL, device, &device_info, 0));
        if (!(device_info.StateFlags & DISPLAY_DEVICE_ATTACHED_TO_DESKTOP)
         || !(device_info.StateFlags & DISPLAY_DEVICE_PRIMARY_DEVICE)) {
            continue;
        }
        DEVMODE default_mode = {.dmSize = sizeof(default_mode)};
        ASSERT(EnumDisplaySettings(device_info.DeviceName,
                                   ENUM_REGISTRY_SETTINGS, &default_mode));
        ASSERT(default_mode.dmFields & DM_PELSWIDTH);
        ASSERT(default_mode.dmFields & DM_PELSHEIGHT);
        *width_ret = default_mode.dmPelsWidth;
        *height_ret = default_mode.dmPelsHeight;
        break;
    }

    ASSERT(*width_ret != -1);

#else
    return 0;
#endif

    return 1;
}

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_graphics_misc)

TEST_INIT(init)
{
    return graphics_init();
}

TEST_CLEANUP(cleanup)
{
    graphics_cleanup();
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_multisample)
{
    if (!(graphics_set_display_attr("multisample", 4))) {
        SKIP("Multisample not supported.");
    }
    if (!open_window(TESTW, TESTH)) {
        SKIP("Unable to set a multisample display mode.");
    }
    graphics_set_viewport(0, 0, TESTW, TESTH);

    graphics_start_frame();

    graphics_clear(0, 0, 0);
    CHECK_TRUE(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){-1,0,0}, NULL, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){-1,1,0}, NULL, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){1.0f/TESTW,1,0}, NULL, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){1.0f/TESTW,0,0}, NULL, NULL));
    CHECK_TRUE(graphics_end_and_draw_primitive());
    uint8_t *pixels;
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        const int x = (i/4) % TESTW;
        const int y = (i/4) / TESTW;
        if (x == TESTW/2 && y >= TESTH/2) {
            /* Sample points are generally pseudorandom, so we don't know
             * exactly what value we'll get back from the renderer, but it
             * should be a non-black, non-white grey. */
            const int p = pixels[i];
            if (p == 0 || p == 255 || pixels[i+1] != p || pixels[i+2] != p
             || pixels[i+3] != 255) {
                FAIL("Pixel (%d,%d) was RGBA (%u,%u,%u,%u) but should have"
                     " been multisampled grey",
                     x, y, pixels[i+0], pixels[i+1], pixels[i+2], pixels[i+3]);
            }
        } else {
            const int p = (x < TESTW/2 && y >= TESTH/2) ? 255 : 0;
            CHECK_PIXEL(&pixels[i], p,p,p,255, x, y);
        }
    }
    mem_free(pixels);

    graphics_finish_frame();
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_grab_after_window_size_change)
{
    if (!graphics_has_windowed_mode()) {
        SKIP("This platform does not support windowed mode.");
    }

    uint8_t *pixels;

    ASSERT(open_window(TESTW, TESTH));
    CHECK_TRUE(graphics_set_display_mode(TESTW*2, TESTH*2, NULL));
    /* Some drivers may need an extra frame before we can read data. */
    graphics_start_frame();
    graphics_finish_frame();
    graphics_start_frame();
    graphics_clear(0.2, 0.4, 0.6);

    ASSERT(pixels = mem_alloc(TESTW * TESTH * 4, 0, MEM_ALLOC_TEMP));
    CHECK_TRUE(graphics_read_pixels(TESTW/2, TESTH/2, TESTW, TESTH, pixels));
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        CHECK_PIXEL(&pixels[i], 51,102,153,255, (i/4) % TESTW, (i/4) / TESTW);
    }
    mem_free(pixels);

    int texture;
    CHECK_TRUE(texture = texture_grab_new(TESTW/2, TESTH/2, TESTW, TESTH,
                                          1, 0, 0));
    CHECK_TRUE(pixels = texture_lock(texture));
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        CHECK_PIXEL(&pixels[i], 51,102,153,255, (i/4) % TESTW, (i/4) / TESTW);
    }
    texture_destroy(texture);

    CHECK_TRUE(texture = texture_create(TESTW, TESTH, 0, 0));
    CHECK_TRUE(texture_grab(texture, TESTW/2, TESTH/2, TESTW, TESTH, 1));
    CHECK_TRUE(pixels = texture_lock(texture));
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        CHECK_PIXEL(&pixels[i], 51,102,153,255, (i/4) % TESTW, (i/4) / TESTW);
    }
    texture_destroy(texture);

    graphics_finish_frame();
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_window_restores_display_mode)
{
    if (!graphics_has_windowed_mode()) {
        SKIP("This platform does not support windowed mode.");
    }

    const int orig_width = graphics_device_width();
    const int orig_height = graphics_device_height();
    int width, height;
    if (!get_alternate_video_mode(&width, &height)) {
        SKIP("No alternate video mode available.");
    }

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(width, height, NULL));
    CHECK_FALSE(graphics_display_is_window());
    CHECK_TRUE(graphics_set_display_attr("window", 1));
    CHECK_TRUE(graphics_set_display_mode(orig_width, orig_height, NULL));
    CHECK_TRUE(graphics_display_is_window());
    int new_width, new_height;
    if (!get_display_resolution(&new_width, &new_height)) {
        FAIL("Don't know how to get display resolution on this platform");
    }
    CHECK_INTEQUAL(new_width, orig_width);
    CHECK_INTEQUAL(new_height, orig_height);

    /* Also check with a same-sized window since that may take a different
     * code path. */
    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(width, height, NULL));
    CHECK_FALSE(graphics_display_is_window());
    CHECK_TRUE(graphics_set_display_attr("window", 1));
    CHECK_TRUE(graphics_set_display_mode(width, height, NULL));
    CHECK_TRUE(graphics_display_is_window());
    if (!get_display_resolution(&new_width, &new_height)) {
        FAIL("Don't know how to get display resolution on this platform");
    }
    CHECK_INTEQUAL(new_width, orig_width);
    CHECK_INTEQUAL(new_height, orig_height);

    return 1;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
