/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/posix/ioqueue.c: Tests for the ioqueue library used by
 * POSIX sys_file_*() functions.
 */

#include "src/base.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/posix/ioqueue.h"
#include "src/sysdep/posix/path_max.h"
#include "src/test/base.h"
#include "src/thread.h"

#define time __UNUSED_time  // Avoid a shadowing warning.
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>
#undef time

/*************************************************************************/
/**************************** Helper routines ****************************/
/*************************************************************************/

/**
 * req_waiter:  Thread wrapper for ioq_wait().
 *
 * [Parameters]
 *     req: Request ID (cast to a pointer).
 * [Return value]
 *     ioq_wait(req)
 */
static int req_waiter(void *req)
{
    return ioq_wait((int)(intptr_t)req, NULL);
}

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

static int do_test_posix_ioqueue(void);
int test_posix_ioqueue(void)
{
#ifdef SIL_PLATFORM_ANDROID
    /* All the data files are stuck in an archive, so open() will fail. */
    SKIP("Skipping ioqueue test on Android (open() not available).");
#endif

    char original_cwd[PATH_MAX];
    ASSERT(getcwd(original_cwd, sizeof(original_cwd)));
    char resource_dir[4096];
    ASSERT(sys_get_resource_path_prefix(resource_dir, sizeof(resource_dir))
           < (int)sizeof(resource_dir));
    ASSERT(chdir(resource_dir) == 0);

    const int result = do_test_posix_ioqueue();

    ASSERT(chdir(original_cwd) == 0);
    return result;
}

/*-----------------------------------------------------------------------*/

DEFINE_GENERIC_TEST_RUNNER(do_test_posix_ioqueue)

TEST_INIT(init)
{
    CHECK_TRUE(thread_init());
    ioq_reset();
    return 1;
}

TEST_CLEANUP(cleanup)
{
    thread_cleanup();
    ioq_reset();
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that files can be opened (on both the first and subsequent
 * requests). */
TEST(test_open)
{
    int req, fd, fd2;
    char buf[5];

    CHECK_TRUE(req = ioq_open("testdata/test.txt", O_RDONLY));
    CHECK_TRUE((fd = ioq_wait(req, NULL)) >= 0);
    CHECK_TRUE(req = ioq_open("testdata/test.txt", O_RDONLY));
    CHECK_TRUE((fd2 = ioq_wait(req, NULL)) >= 0);
    mem_clear(buf, sizeof(buf));
    CHECK_INTEQUAL(read(fd, buf, 5), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);
    mem_clear(buf, sizeof(buf));
    CHECK_INTEQUAL(read(fd2, buf, 5), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);
    close(fd);
    close(fd2);

    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that memory allocation errors while opening files are handled
 * properly. */
TEST(test_open_memory_failure)
{
    int req, fd;
    char buf[5];

    CHECK_MEMORY_FAILURES(
        (req = ioq_open("testdata/test.txt", O_RDONLY))
        && (fd = ioq_wait(req, NULL)) >= 0);
    mem_clear(buf, sizeof(buf));
    CHECK_INTEQUAL(read(fd, buf, 5), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);
    close(fd);

    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that a NULL path is rejected (as opposed to crashing). */
TEST(test_open_null)
{
    CHECK_FALSE(ioq_open(NULL, O_RDONLY));

    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that O_WRONLY and O_RDWR are rejected.  Use /dev/null to avoid
 * accidentally overwriting existing files (but make sure it exists first). */
TEST(test_open_write)
{
    int req, fd;

    CHECK_TRUE(req = ioq_open("/dev/null", O_RDONLY));
    CHECK_TRUE((fd = ioq_wait(req, NULL)) >= 0);
    close(fd);
    CHECK_FALSE(ioq_open("/dev/null", O_WRONLY));
    CHECK_FALSE(ioq_open("/dev/null", O_RDWR));

    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that attempting to open a nonexistent file fails with ENOENT. */
TEST(test_open_noent)
{
    int req, error;

    CHECK_TRUE(req = ioq_open("testdata/no_such_file", O_RDONLY));
    error = 0;
    CHECK_INTEQUAL(ioq_wait(req, &error), -1);
    CHECK_INTEQUAL(error, ENOENT);

    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that files can be read (on both the first and subsequent requests). */
TEST(test_read)
{
    int fd, fd2, req;
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);
    CHECK_TRUE((fd2 = open("testdata/test.txt", O_RDONLY)) >= 0);

    mem_clear(buf, sizeof(buf));
    CHECK_TRUE(req = ioq_read(fd, buf, 5, 0));
    CHECK_INTEQUAL(ioq_wait(req, NULL), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);
    mem_clear(buf, sizeof(buf));
    CHECK_TRUE(req = ioq_read(fd2, buf, 5, 0));
    CHECK_INTEQUAL(ioq_wait(req, NULL), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);

    close(fd);
    close(fd2);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that the testing-specific flag TEST_posix_ioqueue_block_io_thread
 * properly blocks I/O. */
TEST(test_block_io_thread)
{
    int fd, req;
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);

    mem_clear(buf, sizeof(buf));
    TEST_posix_ioqueue_block_io_thread(1);
    CHECK_TRUE(req = ioq_read(fd, buf, 5, 0));
    /* Wait long enough that we can be reasonably sure the data would have
     * been loaded if the thread was running. */
    nanosleep(&(struct timespec){0, 100000000}, NULL);  // 0.1 sec
    CHECK_FALSE(ioq_poll(req));
    TEST_posix_ioqueue_block_io_thread(0);
    CHECK_INTEQUAL(ioq_wait(req, NULL), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);

    close(fd);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that the testing-specific flag TEST_posix_ioqueue_unblock_on_wait
 * properly unblocks I/O on an ioq_wait() call. */
TEST(test_unblock_on_wait)
{
    int fd, req;
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);

    mem_clear(buf, sizeof(buf));
    TEST_posix_ioqueue_block_io_thread(1);
    TEST_posix_ioqueue_unblock_on_wait(1);
    CHECK_TRUE(req = ioq_read(fd, buf, 5, 0));
    CHECK_INTEQUAL(ioq_wait(req, NULL), 5);  // Will freeze on test failure.
    CHECK_MEMEQUAL(buf, "hello", 5);
    TEST_posix_ioqueue_unblock_on_wait(0);

    close(fd);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that the testing-specific flag TEST_posix_ioqueue_permfail_next_read
 * properly fails a single read request. */
TEST(test_permfail_next_read)
{
    int fd, fd2, req;
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);
    CHECK_TRUE((fd2 = open("testdata/test.txt", O_RDONLY)) >= 0);

    TEST_posix_ioqueue_permfail_next_read(1);
    mem_clear(buf, sizeof(buf));
    CHECK_FALSE(req = ioq_read(fd, buf, 5, 0));
    CHECK_INTEQUAL(errno, ENOMEM);

    mem_clear(buf, sizeof(buf));
    CHECK_TRUE(req = ioq_read(fd2, buf, 5, 0));
    CHECK_INTEQUAL(ioq_wait(req, NULL), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);

    close(fd);
    close(fd2);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that the testing-specific flag TEST_posix_ioqueue_tempfail_next_read
 * properly fails a single read request. */
TEST(test_tempfail_next_read)
{
    int fd, fd2, req;
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);
    CHECK_TRUE((fd2 = open("testdata/test.txt", O_RDONLY)) >= 0);

    TEST_posix_ioqueue_tempfail_next_read(1);
    mem_clear(buf, sizeof(buf));
    CHECK_FALSE(req = ioq_read(fd, buf, 5, 0));
    CHECK_INTEQUAL(errno, EAGAIN);

    mem_clear(buf, sizeof(buf));
    CHECK_TRUE(req = ioq_read(fd2, buf, 5, 0));
    CHECK_INTEQUAL(ioq_wait(req, NULL), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);

    close(fd);
    close(fd2);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that the testing-specific flag TEST_posix_ioqueue_iofail_next_read
 * properly fails a single read request. */
TEST(test_iofail_next_read)
{
    int fd, fd2, req, error;
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);
    CHECK_TRUE((fd2 = open("testdata/test.txt", O_RDONLY)) >= 0);

    TEST_posix_ioqueue_iofail_next_read(1);
    mem_clear(buf, sizeof(buf));
    CHECK_TRUE(req = ioq_read(fd, buf, 5, 0));
    CHECK_INTEQUAL(ioq_wait(req, &error), -1);
    CHECK_INTEQUAL(error, EIO);
    CHECK_MEMEQUAL(buf, "\0\0\0\0\0", 5);

    mem_clear(buf, sizeof(buf));
    CHECK_TRUE(req = ioq_read(fd2, buf, 5, 0));
    CHECK_INTEQUAL(ioq_wait(req, NULL), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);

    close(fd);
    close(fd2);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that memory allocation errors while reading files are handled
 * properly. */
TEST(test_read_memory_failure)
{
    int fd, req, res;
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);

    mem_clear(buf, sizeof(buf));
    CHECK_MEMORY_FAILURES(
        (req = ioq_read(fd, buf, 5, 0))
        && (res = ioq_wait(req, NULL)) >= 0);
    CHECK_INTEQUAL(res, 5);
    CHECK_MEMEQUAL(buf, "hello", 5);

    close(fd);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that invalid read parameters are rejected. */
TEST(test_read_invalid)
{
    int fd;
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);

    CHECK_FALSE(ioq_read(-1, buf, 5, 0));
    CHECK_FALSE(ioq_read(fd, NULL, 5, 0));
    CHECK_FALSE(ioq_read(fd, buf, -1, 0));
    CHECK_FALSE(ioq_read(fd, buf, 5, -1));

    close(fd);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that reads executed in parallel work properly. */
TEST(test_read_parallel)
{
    int fd, fd2, req, req2;
    char buf[5], buf2[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);
    CHECK_TRUE((fd2 = open("testdata/test.txt", O_RDONLY)) >= 0);

    mem_clear(buf, sizeof(buf));
    mem_clear(buf2, sizeof(buf2));
    CHECK_TRUE(req = ioq_read(fd, buf, 5, 0));
    CHECK_TRUE(req2 = ioq_read(fd2, buf2, 5, 0));
    CHECK_INTEQUAL(ioq_wait(req, NULL), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);
    CHECK_INTEQUAL(ioq_wait(req2, NULL), 5);
    CHECK_MEMEQUAL(buf2, "hello", 5);

    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check handling of memory allocation failures for parallel reads. */
TEST(test_read_parallel_memory_failure)
{
    int fd, fd2, req, req2, res2;
    char buf[5], buf2[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);
    CHECK_TRUE((fd2 = open("testdata/test.txt", O_RDONLY)) >= 0);

    mem_clear(buf, sizeof(buf));
    mem_clear(buf2, sizeof(buf2));
    CHECK_TRUE(req = ioq_read(fd, buf, 5, 0));
    CHECK_MEMORY_FAILURES(
        (req2 = ioq_read(fd2, buf2, 5, 0))
        && (res2 = ioq_wait(req2, NULL)) >= 0);
    CHECK_INTEQUAL(ioq_wait(req, NULL), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);
    CHECK_INTEQUAL(res2, 5);
    CHECK_MEMEQUAL(buf2, "hello", 5);

    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check handling of memory allocation failures for parallel reads when
 * a shrinking realloc() call fails. */
TEST(test_read_parallel_realloc_failure)
{
    int fd, fd2, req, req2;
    char buf[5], buf2[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);
    CHECK_TRUE((fd2 = open("testdata/test.txt", O_RDONLY)) >= 0);

    mem_clear(buf, sizeof(buf));
    mem_clear(buf2, sizeof(buf2));
    CHECK_TRUE(req = ioq_read(fd, buf, 5, 0));
    /* CHECK_MEMORY_FAILURES() would report a leak when the shrinking
     * realloc call fails, so we do this manually and rely on the
     * automatic leak check when the test completes. */
    for (int i = 0; ; i++) {
        if (i >= 100) {
            FAIL("ioq_read(fd2, buf2, 5, 0) did not succeed after 100"
                 " iterations");
        }
        TEST_mem_fail_after(i, 1, 1);
        req2 = ioq_read(fd2, buf2, 5, 0);
        TEST_mem_fail_after(-1, 0, 0);
        if (req2 != 0) {
            break;
        }
    }
    CHECK_INTEQUAL(ioq_wait(req2, NULL), 5);
    CHECK_INTEQUAL(ioq_wait(req, NULL), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);
    CHECK_MEMEQUAL(buf2, "hello", 5);

    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that overlapping requests work if the request array is expanded
 * while a request is being waited on. */
TEST(test_read_parallel_2)
{
    int fd, fd2, req, req2;
    char buf[5], buf2[5];
    int thread;

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);
    CHECK_TRUE((fd2 = open("testdata/test.txt", O_RDONLY)) >= 0);

    TEST_posix_ioqueue_move_on_realloc(1);
    TEST_posix_ioqueue_block_io_thread(1);
    mem_clear(buf, sizeof(buf));
    mem_clear(buf2, sizeof(buf2));
    CHECK_TRUE(req = ioq_read(fd, buf, 5, 0));
    ASSERT(thread = thread_create(req_waiter, (void *)(intptr_t)req));
    /* Wait long enough that we can be reasonably sure the thread is waiting
     * on the request. */
    nanosleep(&(struct timespec){0, 1000000}, NULL);  // 1 msec
    CHECK_TRUE(req2 = ioq_read(fd2, buf2, 5, 0));
    TEST_posix_ioqueue_block_io_thread(0);
    /* Check req2 first; if it's finished, then req is guaranteed to have
     * finished as well, and (in theory) we shouldn't have to wait below. */
    CHECK_INTEQUAL(ioq_wait(req2, NULL), 5);
    CHECK_MEMEQUAL(buf2, "hello", 5);
    int wait_ms = 1000;
    while (thread_is_running(thread)) {
        if (wait_ms <= 0) {
            FAIL("ioq_wait(req) did not return");
        }
        nanosleep(&(struct timespec){0, 1000000}, NULL);  // 1 msec
        wait_ms--;
    }
    CHECK_INTEQUAL(thread_wait(thread), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);

    TEST_posix_ioqueue_move_on_realloc(0);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that the request array is shrunk when there are a lot of unused
 * entries. */
TEST(test_shrink_request_array)
{
    int fd, req[20];
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);
    const uint64_t used1 = mem_debug_bytes_allocated();

    for (int i = 0; i < lenof(req); i++) {
        CHECK_TRUE(req[i] = ioq_read(fd, buf, 5, 0));
    }
    const uint64_t used2 = mem_debug_bytes_allocated();

    for (int i = lenof(req) - 1; i >= 0; i--) {
        CHECK_INTEQUAL(ioq_wait(req[i], NULL), 5);
    }
    const uint64_t used3 = mem_debug_bytes_allocated();

    CHECK_TRUE(used3 < used2);
    CHECK_TRUE(used3 > used1); // There should still be some entries allocated.

    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check memory reallocation failure when shrinking the request array. */
TEST(test_shrink_request_array_memory_failure)
{
    int fd, req[20];
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);

    for (int i = 0; i < lenof(req); i++) {
        CHECK_TRUE(req[i] = ioq_read(fd, buf, 5, 0));
    }
    for (int i = lenof(req) - 1; i >= 0; i--) {
        CHECK_INTEQUAL(ioq_wait(req[i], NULL), 5);
    }
    const uint64_t used1 = mem_debug_bytes_allocated();

    for (int i = 0; i < lenof(req); i++) {
        CHECK_TRUE(req[i] = ioq_read(fd, buf, 5, 0));
    }
    for (int i = lenof(req) - 1; i >= 0; i--) {
        TEST_mem_fail_after(0, 1, 1);
        const int res = ioq_wait(req[i], NULL);
        TEST_mem_fail_after(-1, 0, 0);
        CHECK_INTEQUAL(res, 5);
    }
    const uint64_t used2 = mem_debug_bytes_allocated();

    CHECK_TRUE(used2 > used1);

    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check ioq_poll() and ioq_cancel() behavior. */
TEST(test_poll_cancel)
{
    int fd, req, error;
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);

    /* Check that ioq_poll() returns true at some point for a valid request
     * (which is about the best we can do for testing). */
    CHECK_TRUE(req = ioq_read(fd, buf, 5, 0));
    while (!ioq_poll(req)) {
        thread_yield();
    }
    CHECK_INTEQUAL(ioq_wait(req, NULL), 5);
    CHECK_MEMEQUAL(buf, "hello", 5);

    /* Check that ioq_poll() returns true for invalid request IDs. */
    CHECK_TRUE(ioq_poll(0));
    CHECK_TRUE(ioq_poll(INT_MAX));
    CHECK_TRUE(ioq_poll(req));

    /* Check that ioq_wait() fails for invalid request IDs. */
    errno = 0;
    CHECK_INTEQUAL(ioq_wait(0, NULL), -1);
    CHECK_INTEQUAL(errno, ESRCH);
    errno = 0;
    CHECK_INTEQUAL(ioq_wait(INT_MAX, NULL), -1);
    CHECK_INTEQUAL(errno, ESRCH);
    errno = 0;
    error = 0;
    CHECK_INTEQUAL(ioq_wait(req, &error), -1);
    CHECK_INTEQUAL(errno, ESRCH);
    CHECK_INTEQUAL(error, ESRCH);

    close(fd);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that ioq_cancel() properly cancels requests. */
TEST(test_cancel)
{
    int fd, req, req2, error;
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);

    CHECK_TRUE(req = ioq_open("testdata/nonexistent.txt", O_RDONLY));
    CHECK_TRUE(req2 = ioq_read(fd, buf, 5, 0));
    while (!ioq_poll(req2)) {
        thread_yield();
    }
    ioq_cancel(req);
    ioq_cancel(req2);
    error = 0;
    CHECK_INTEQUAL(ioq_wait(req, &error), -1);
    CHECK_INTEQUAL(error, ECANCELED);
    error = 0;
    CHECK_INTEQUAL(ioq_wait(req2, &error), -1);
    CHECK_INTEQUAL(error, ECANCELED);

    close(fd);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that ioq_cancel() properly cancels a request that has not yet
 * begun. */
TEST(test_cancel_pending_request)
{
    int fd, req, error;
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);
    mem_clear(buf, 5);

    TEST_posix_ioqueue_block_io_thread(1);
    CHECK_TRUE(req = ioq_read(fd, buf, 5, 0));
    ioq_cancel(req);
    TEST_posix_ioqueue_block_io_thread(0);
    error = 0;
    CHECK_INTEQUAL(ioq_wait(req, &error), -1);
    CHECK_INTEQUAL(error, ECANCELED);
    CHECK_MEMEQUAL(buf, "\0\0\0\0\0", 5);

    close(fd);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that ioq_cancel() on a completed open operation closes the FD. */
TEST(test_cancel_open)
{
    int fd, req, error;

    /* We assume that repeated sequences of open() followed by close() will
     * always return the same FD, and see if there's any change in FD from
     * a raw open() call before and after a cancelled open operation. */

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);
    close(fd);
    CHECK_TRUE(open("testdata/test.txt", O_RDONLY) == fd);
    close(fd);
    CHECK_TRUE(req = ioq_open("testdata/test.txt", O_RDONLY));
    while (!ioq_poll(req)) {
        thread_yield();
    }
    ioq_cancel(req);
    error = 0;
    CHECK_INTEQUAL(ioq_wait(req, &error), -1);
    CHECK_INTEQUAL(error, ECANCELED);
    CHECK_INTEQUAL(open("testdata/test.txt", O_RDONLY), fd);

    /* Check that ioq_cancel() doesn't crash on an invalid request ID. */
    ioq_cancel(0);
    ioq_cancel(req);

    close(fd);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that ioq_reset() properly cancels pending requests.  We can't
 * check a cancelled request's status directly, so we use the same test as
 * for cancelled open operations. */
TEST(test_reset)
{
    int fd, fd2, req, req2;
    char buf[5];

    CHECK_TRUE((fd = open("testdata/test.txt", O_RDONLY)) >= 0);
    CHECK_TRUE((fd2 = open("testdata/test.txt", O_RDONLY)) >= 0);
    close(fd2);
    CHECK_TRUE(req = ioq_open("testdata/nonexistent.txt", O_RDONLY));
    CHECK_TRUE(req2 = ioq_open("testdata/test.txt", O_RDONLY));
    while (!ioq_poll(req2)) {
        thread_yield();
    }
    TEST_posix_ioqueue_block_io_thread(1);
    CHECK_TRUE(ioq_read(fd, buf, 5, 0));
    CHECK_TRUE(ioq_open("testdata/nonexistent.txt", O_RDONLY));
    TEST_posix_ioqueue_block_io_thread(0);
    ioq_reset();
    CHECK_INTEQUAL(open("testdata/test.txt", O_RDONLY), fd2);

    close(fd);
    close(fd2);
    return 1;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
